#include "burp.h"
#include "prog.h"
#include "msg.h"
#include "rs_buf.h"
#include "lock.h"
#include "handy.h"
#include "asyncio.h"
#include "counter.h"
#include "client_vss.h"
#include "auth_client.h"
#include "backup_phase1_client.h"
#include "backup_phase2_client.h"
#include "restore_client.h"
#include "list_client.h"
#include "ssl.h"

static int init_socket(const char *host, int port)
{
        int rfd=-1;
        struct sockaddr_in sin;
        struct hostent *hp=NULL;

        if(!(hp=gethostbyname(host)))
        {
                logp("unknown host: %s\n", host);
                return -1;
        }

        memset(&sin, 0, sizeof(sin));
        sin.sin_family=AF_INET;
        memcpy((char *)&sin.sin_addr, hp->h_addr, hp->h_length);
        sin.sin_port=htons(port);

        if((rfd=socket(PF_INET, SOCK_STREAM, 0))<0)
        {
		berrno be;
                logp("socket error: %s\n", be.bstrerror());
                return -1;
        }
        //setsockopt(rfd, IPPROTO_TCP, TCP_NODELAY, (char *) &flag, sizeof(int));
        if(connect(rfd, (struct sockaddr *)&sin, sizeof(sin))<0)
        {
		//berrno be;
                logp("connect error, %s: %s\n", host, strerror(errno));
			//be.bstrerror(b_errno_win32));
                close_fd(&rfd);
                return -1;
        }

#ifdef HAVE_WIN32
	setmode(rfd, O_BINARY);
#endif
        return rfd;
}

static int do_backup_client(struct config *conf)
{
	int ret=0;

logp("do backup client\n");

#if defined(HAVE_WIN32)
	win32_enable_backup_privileges(1 /* ignore_errors */);
#endif
#if defined(WIN32_VSS)
	win32_start_vss(conf);
#endif

	// Scan the file system and send the results to the server.
	if(!ret) ret=backup_phase1_client(conf);

	// Now, the server will be telling us what data we need to send.
	if(!ret) ret=backup_phase2_client();

#if defined(WIN32_VSS)
	win32_stop_vss();
#endif

	if(ret) logp("error in backup\n");
	else logp("backup finished ok\n");

	return ret;
}

/*
static void sighandler(int sig)
{
	logp("got signal: %d\n", sig);
	// Close the sockets properly so as to avoid annoying waits during
	// testing when I kill the server with a Ctrl-C and then get
	// 'unable to bind listening socket'.
	async_free();
	logp("exiting\n");
	exit(1);
}
*/

static void setup_signals(void)
{
	// Ignore SIGPIPE - we are careful with read and write return values.
#ifndef HAVE_WIN32
	//signal(SIGABRT, &sighandler);
	//signal(SIGTERM, &sighandler);
	//signal(SIGINT, &sighandler);
#endif
}

static int s_server_session_id_context=1;

int client(struct config *conf, enum action act, const char *backup, const char *restoreprefix, const char *regex, int forceoverwrite)
{
	int ret=0;
	int rfd=-1;
	SSL *ssl=NULL;
	BIO *sbio=NULL;
	SSL_CTX *ctx=NULL;

	setup_signals();
//	settimers(0, 100);
	logp("begin client\n");

	ssl_load_globals();
	if(!(ctx=ssl_initialise_ctx(conf)))
	{
		logp("error initialising ssl ctx\n");
		return -1;
	}

	SSL_CTX_set_session_id_context(ctx,
		(const unsigned char *)&s_server_session_id_context,
		sizeof(s_server_session_id_context));

	if((rfd=init_socket(conf->server, conf->port))<0) return 1;

	if(!(ssl=SSL_new(ctx))
	  || !(sbio=BIO_new_socket(rfd, BIO_NOCLOSE)))
	{
		logp("There was a problem joining ssl to the socket.\n");
		close_fd(&rfd);
		return 1;
	}
	SSL_set_bio(ssl, sbio, sbio);
	if(SSL_connect(ssl)<=0)
	{
		berr_exit("SSL connect error\n");
		close_fd(&rfd);
		return 1;
	}
	if(ssl_check_cert(ssl, conf))
	{
		logp("check cert failed\n");
		close_fd(&rfd);
		return 1;
	}
	set_non_blocking(rfd);

	if(!(ret=async_init(rfd, ssl))
	 && !(ret=authorise_client(conf)))
	{
		rfd=-1;
		switch(act)
		{
			case ACTION_BACKUP:
				ret=do_backup_client(conf);
				break;
			case ACTION_RESTORE:
			case ACTION_VERIFY:
				ret=do_restore_client(conf, act, backup,
					restoreprefix, regex, forceoverwrite);
				break;
			case ACTION_LIST:
			case ACTION_LONG_LIST:
			default:
				ret=do_list_client(backup, regex, act);
				break;
		}
	}

	rfd=-1;
	async_free();
	ssl_destroy_ctx(ctx);

        //logp("end client\n");
	return ret;
}
